#include <iostream>
#include <verilated.h>
#include "VCAMT8D32.h"

using namespace std;

// Current simulation time (64-bit unsigned)
vluint64_t main_time = 0;
vluint64_t max_time = 200;

// Called by $time in Verilog
double sc_time_stamp() {
    return main_time;  // Note does conversion to real, to match SystemC
}

int main(int argc, char** argv, char** env){
  
  // Prevent unused variable warnings
  if (false && argc && argv && env) {}

  // Set debug level, 0 is off, 9 is highest presently used
  // May be overridden by commandArgs
  Verilated::debug(0);

  // Randomization reset policy
  // May be overridden by commandArgs
  Verilated::randReset(2);

  // Verilator must compute traced signals
  Verilated::traceEverOn(true);

  // Pass arguments so Verilated code can see them, e.g. $value$plusargs
  // This needs to be called before you create any model
  Verilated::commandArgs(argc, argv);

  // Create logs/ directory in case we have traces to put under it
  Verilated::mkdir("logs");

  // Construct the Verilated model, from Vtop.h generated from Verilating "top.v"
  VCAMT8D32* top = new VCAMT8D32();

  top->clk = 0;
  top->res_n = !0;

  top->lock = 0;
  top->unlock = 0;
  
  top->wr = 0;
  top->din = 0xAAAAAAAAAA;
    
  top->rd = 0;

  while (!Verilated::gotFinish()) {

    main_time++;

    // toggle clock
    top->clk=!top->clk;

    if(!top->clk){
    
      // Toggle control signals on an edge that doesn't correspond
      // to where the controls are sampled; in this example we do
      // this only on a negedge of clk, because we know
      // reset is not sampled there.
      if ( main_time > 0 && main_time < 6 ) {
	top->res_n = !1;  //   Assert reset
      } else {
	top->res_n = !0;  // Deassert reset
      }

      if(main_time == 6)
	top->din = 0xAAAAAAAAAA;

      // UNLOCK CAM
      //----------------
      if( main_time>6 && main_time<=8 ) {
	// unlock CAMT8D32
	top->key = 0xBABE;
	top->unlock = 1;
      }

      // WRITE TEST DATA
      //----------------
      if (main_time>8 && main_time<floor(max_time/4)) {
	top->unlock = 0;
	top->key = 0x00000000;
	if (0==main_time%20) {
	  // increment or wrap data
	  if(top->din==0xFFFFFFFFFF) {
	    top->din=0x0000000000;
	  } else {
	    top->din += 0x0000000001;
	  }
	}
	if (0==main_time%2) {
	  // increment or wrap data
	  if(top->din==0xFFFFFFFFFF) {
	    top->din=0x0000000000;
	  } else {
	    top->din += 0x0100000000;
	  }
	}
	if (0==main_time%4) {
	  // write to CAMT8D32
	  top->wr = 1;	    
	} else {
	  top->wr = 0;
	}
      }

      // READ TEST DATA
      //----------------
      if(main_time==floor(max_time/4)){
	  top->din = 0xAA00000000;
	  top->wr = 0;
      }

	if (main_time>floor(max_time/4)+2 && main_time<floor(max_time/2)) {
	  // read from CAMT8D32
	  top->wr = 0;
	  if (0==main_time%4)
	    top->din += 0x0100000000;
	  if (0==main_time%4) {
	  // read from CAMT8D32
	    top->rd = 1;
	  } else {
	    top->rd = 0;
	  }
	}

	if(main_time==floor(max_time/2))
	  top->din = 0xAAAAAAAAAA;	  
      
      
	if (main_time>floor(max_time/2) && main_time<=floor(4*max_time/6)) {
	  
	  // lock stack
	  if(main_time==floor(max_time/2)+2)
	    top->lock = 1;
	  else
	    top->lock = 0;

	  // and read from locked stack
	  if (0==main_time%4) {
	    top->rd = 1;
	  } else {
	    top->rd = 0;
	  }
	}
      
	if (main_time>floor(4*max_time/6) && main_time<=floor(5*max_time/6)) {

	  // lock stack
	  if(main_time==floor(max_time/2)+2) top->lock = 1; else top->lock = 0;

	  // toggle din
	  if (0==main_time%4) {
	    // increment or wrap data
	    if(top->din==0xFFFFFFFFFF) {
	      top->din=0x0000000000;
	    } else {
	      top->din += 0x010000000F;
	    }
	  }

	  // and write to locked stack
	  if (0==main_time%4) {
	    top->wr = 1;
	  } else {
	    top->wr = 0;
	  }
	}
	
	if (main_time>floor(5*max_time/6) && main_time<=floor(max_time)) {

	  // unlock stack
	  if(main_time==floor(5*max_time/6)+2){
	    top->lock = 0;
	    top->unlock = 1;
	    top->key = 0xBABE;
	  } else {
	    top->lock = 0;
	    top->unlock = 0;
	    top->key = 0xBABE;
	  }

	  if (0==main_time%4) {
	    // increment or wrap data
	    if(top->din==0xFFFFFFFFFF) {
	      top->din=0x0000000000;
	    } else {
	      top->din += 0x01000000000F;
	    }
	  }

	  // concurrent read write access
	  if (0==main_time%4) {
	    top->wr = 1;
	    top->rd = 1;
	  } else {
	    top->wr = 0;
	    top->rd = 1;
	  }
	}      

      if (main_time>=max_time) {	
	top->wr=1; top->lock=1; top->unlock=1; top->res_n=0;
      }
	
    }
       
    // Evaluate model
    // (If you have multiple models being simulated in the same
    // timestep then instead of eval(), call eval_step() on each, then
    // eval_end_step() on each.)
    top->eval();
    
    if(!top->clk){
      // Read outputs
      VL_PRINTF( "[%" VL_PRI64 "d] clk=%x res_n=%x lock=%x ( unlock=%x key=%4x ) ( wr=%x din=%10lx ) rd=%x => locked=%x ( dout_val=%x, dout=%10lx )\n",
		 main_time,
		 top->clk, top->res_n,
		 top->lock, top->unlock, top->key,
		 top->wr, top->din, top->rd,
		 top->locked, top->dout_val, top->dout );             
      }
  }
  
  top->final();

  //  Coverage analysis (since test passed)
#if VM_COVERAGE
  Verilated::mkdir("logs");
  VerilatedCov::write("logs/coverage.dat");
#endif

  // Delete model
  delete top;
  top = nullptr;

  // Exit
  exit(0);
  
}
